"""The client for the Dummy Executor."""

from typing import Dict, List, Tuple, Union

from jobmon.core.cluster_protocol import ClusterQueue


class DummyQueue(ClusterQueue):
    """Implementation of the dummy executor queue, derived from ClusterQueue."""

    def __init__(self, queue_id: int, queue_name: str, parameters: dict) -> None:
        """Intialization of SequentialQueue.

        Get the limits from database in client.
        """
        self._queue_id = queue_id
        self._queue_name = queue_name
        self._parameters = parameters

    def validate_resources(
        self, strict: bool = False, **kwargs: Union[str, int, float]
    ) -> Tuple[bool, str]:
        """No resources defined for dummy execution. All resources valid."""
        return True, ""

    def coerce_resources(self, **kwargs: Union[str, int, float]) -> Dict:
        return kwargs

    @property
    def queue_id(self) -> int:
        """Return the ID of the queue."""
        return self._queue_id

    @property
    def queue_name(self) -> str:
        """Return the name of the queue."""
        return self._queue_name

    @property
    def parameters(self) -> Dict:
        """Return the dictionary of parameters."""
        return self._parameters

    @property
    def required_resources(self) -> List:
        """No specified required resources for dummy, return empty list."""
        return []



"""Constants declared for different statuses, types and codes throughout Jobmon."""

from enum import Enum


class ArgType:
    NODE_ARG = 1
    TASK_ARG = 2
    OP_ARG = 3


class TaskResourcesType:
    """Constant Types for Task Resources."""

    ORIGINAL = "O"
    VALIDATED = "V"
    ADJUSTED = "A"


class TaskInstanceStatus:
    """Statuses used for Task Instances."""

    DONE = "D"
    ERROR = "E"
    ERROR_FATAL = "F"
    INSTANTIATED = "I"
    KILL_SELF = "K"
    LAUNCHED = "O"
    QUEUED = "Q"
    RUNNING = "R"
    TRIAGING = "T"

    UNKNOWN_ERROR = "U"
    NO_DISTRIBUTOR_ID = "W"
    NO_HEARTBEAT = "X"
    RESOURCE_ERROR = "Z"


class TaskStatus:
    """Statuses used for Tasks."""

    REGISTERING = "G"
    QUEUED = "Q"
    INSTANTIATING = "I"
    LAUNCHED = "O"
    RUNNING = "R"
    DONE = "D"
    ERROR_RECOVERABLE = "E"
    ADJUSTING_RESOURCES = "A"
    ERROR_FATAL = "F"

    LABEL_DICT = {
        "G": "REGISTERING",
        "Q": "QUEUED",
        "I": "INSTANTIATING",
        "O": "LAUNCHED",
        "R": "RUNNING",
        "D": "DONE",
        "E": "ERROR_RECOVERABLE",
        "A": "ADJUSTING_RESOURCES",
        "F": "ERROR_FATAL",
    }


class WorkflowRunStatus:
    """Statuses used for Workflow Runs."""

    REGISTERED = "G"
    LINKING = "L"
    BOUND = "B"
    ABORTED = "A"
    RUNNING = "R"
    DONE = "D"
    STOPPED = "S"
    ERROR = "E"
    COLD_RESUME = "C"
    HOT_RESUME = "H"
    TERMINATED = "T"
    INSTANTIATED = "I"
    LAUNCHED = "O"


class WorkflowStatus:
    """Statuses used for Workflows."""

    REGISTERING = "G"
    QUEUED = "Q"
    ABORTED = "A"
    INSTANTIATING = "I"
    LAUNCHED = "O"
    RUNNING = "R"
    DONE = "D"
    HALTED = "H"
    FAILED = "F"


class Direction(Enum):
    """A generic utility class.

    Used to represent one-dimensional direction,
    such as upstream/downstream.
    """

    UP = "up"
    DOWN = "down"


class SpecialChars:
    """A generic utility class.

    Used to define special chars.
    """

    ILLEGAL_SPECIAL_CHARACTERS = r"/\\'\" "


class ExecludeTTVs:
    """A hard-coded list.

    Used to exclude task template versions with huge tasks that cause DB crash.
    """

    EXECLUDE_TTVS = {1}  # bashtask


class MaxConcurrentlyRunning:
    """A hard limit of array concurrency.

    Currently set to max int, but can change as INFRA requests.
    """

    import sys

    MAXCONCURRENTLYRUNNING = 2_147_483_647  # mysql int(11) max



"""Multiprocess executes tasks in parallel if multiple threads are available."""

import logging
from multiprocessing import JoinableQueue, Process, Queue
import os
import queue
import shutil
import subprocess
from typing import Any, Dict, List, Optional, Set, Tuple

import psutil

from jobmon.core.cluster_protocol import ClusterDistributor, ClusterWorkerNode
from jobmon.core.constants import TaskInstanceStatus
from jobmon.core.exceptions import RemoteExitInfoNotAvailable

logger = logging.getLogger(__name__)


class PickableTask:
    """Object passed between processes."""

    def __init__(
        self,
        distributor_id: str,
        command: str,
        task_type: str = "array",
    ) -> None:
        """Initialization of PickableTask.

        array_step_id: is only meaningful and has int value when for array
        """
        self.distributor_id = distributor_id
        self.command = command
        self.task_type = task_type


class Consumer(Process):
    """Consumes the tasks to be run."""

    def __init__(self, task_queue: JoinableQueue, response_queue: Queue) -> None:
        """Consume work sent from LocalExecutor through multiprocessing queue.

        this class is structured based on
        https://pymotw.com/2/multiprocessing/communication.html

            task_queue:
                a (multiprocessing.JoinableQueue[Optional[PickableTask]]) object
                created by LocalExecutor used to retrieve work from the
                distributor.
            response_queue:
                A (Queue[Tuple[int, Optional[int], Optional[int]]]) object,
                that will hold information with Queue:
                Tuple[distributor_id, array_step_id if applicable, pid]
        """
        super().__init__()

        # consumer communication
        self.task_queue: JoinableQueue[Optional[PickableTask]] = task_queue
        self.response_queue: Queue[Tuple[str, Optional[int]]] = response_queue

    def run(self) -> None:
        """Wait for work, the execute it."""
        logger.info(f"consumer alive. pid={os.getpid()}")

        while True:
            try:
                task = self.task_queue.get(timeout=1)
                if task is None:
                    logger.info("Received poison pill. Shutting down")
                    # Received poison pill, no more tasks to run
                    self.task_queue.task_done()
                    break

                else:
                    logger.info(f"consumer received {task.command}")

                    # run the job
                    env = os.environ.copy()

                    if task.task_type == "array":
                        job_id, array_step_id = task.distributor_id.split("_")
                        env["JOB_ID"] = job_id
                        env["ARRAY_STEP_ID"] = array_step_id
                    else:
                        env["JOB_ID"] = task.distributor_id

                    proc = subprocess.Popen(task.command, env=env, shell=True)

                    # log the pid with the distributor class
                    self.response_queue.put((task.distributor_id, proc.pid))

                    # wait till the process finishes
                    proc.communicate()

                    logger.info(f"consumer finished processing {task.distributor_id}")

                    # tell the queue this job is done so it can be shut down
                    # someday
                    self.response_queue.put((task.distributor_id, None))
                    self.task_queue.task_done()

            except queue.Empty:
                pass
            except Exception as e:
                logger.exception(e)


class MultiprocessDistributor(ClusterDistributor):
    """Executes tasks locally in parallel.

    It uses the multiprocessing Python library and queues to parallelize the execution of
    tasks. The subprocessing pattern looks like this:
        LocalExec
        --> consumer1
        ----> subconsumer1
        --> consumer2
        ----> subconsumer2
        ...
        --> consumerN
        ----> subconsumerN
    """

    def __init__(
        self, cluster_name: str, parallelism: int = 3, *args: tuple, **kwargs: dict
    ) -> None:
        """Initialization of the multiprocess distributor.

        Args:
            cluster_name: the name of the cluster.
            parallelism (int, optional): how many parallel jobs to distribute at a
                time
        """
        self.temp_dir: Optional[str] = None
        self.started = False
        self._cluster_name = cluster_name

        worker_node_entry_point = shutil.which("worker_node_entry_point")
        if not worker_node_entry_point:
            raise ValueError("worker_node_entry_point can't be found.")
        self._worker_node_entry_point = worker_node_entry_point

        logger.info("Initializing {}".format(self.__class__.__name__))

        self._parallelism = parallelism
        self._next_job_id = 1

        # mapping of Tuple[distributor_id, optinal array_step_id] to pid.
        # if pid is None then it is queued
        self._running_or_submitted: Dict[str, Optional[int]] = {}

        # ipc queues
        self.task_queue: JoinableQueue[Optional[PickableTask]] = JoinableQueue()
        self.response_queue: Queue[Tuple[str, Optional[int]]] = Queue()

        # workers
        self.consumers: List[Consumer] = []

    @property
    def worker_node_entry_point(self) -> str:
        """Path to jobmon worker_node_entry_point."""
        return self._worker_node_entry_point

    @property
    def cluster_name(self) -> str:
        """Return the name of the cluster type."""
        return self._cluster_name

    def _get_subtask_id(self, distributor_id: int, array_step_id: int) -> str:
        """Get the subtask_id based on distributor_id and array_step_id."""
        return str(distributor_id) + "_" + str(array_step_id)

    def start(self) -> None:
        """Fire up N task consuming processes using Multiprocessing.

        Number of consumers is controlled by parallelism.
        """
        # set jobmon command if provided
        if not self.started:
            self.consumers = [
                Consumer(task_queue=self.task_queue, response_queue=self.response_queue)
                for i in range(self._parallelism)
            ]
            for w in self.consumers:
                w.start()

            """Start the default."""
            self.started = True

    def stop(self) -> None:
        """Terminate consumers and call sync 1 final time."""
        actual = self.get_submitted_or_running()
        self.terminate_task_instances(list(actual))

        # Sending poison pill to all worker
        for _ in self.consumers:
            self.task_queue.put(None)

        # Wait for commands to finish
        self.task_queue.join()

        self.started = False

    def _update_internal_states(self) -> None:
        while not self.response_queue.empty():
            distributor_id, pid = self.response_queue.get()
            if pid is not None:
                self._running_or_submitted.update({distributor_id: pid})
            else:
                self._running_or_submitted.pop(distributor_id)

    def terminate_task_instances(self, distributor_ids: List[str]) -> None:
        """Terminate task instances.

        Only terminate the task instances that are running, not going to kill the jobs that
        are actually still in a waiting or a transitioning state.

        Args:
            distributor_ids: A list of distributor IDs.
        """
        logger.debug(f"Going to terminate: {distributor_ids}")

        # first drain the work queue so there are no race conditions with the
        # workers
        current_work: List[Optional[PickableTask]] = []
        work_order: Dict[int, PickableTask] = {}
        dist_ids_work_order: Set[str] = set()
        i = 0
        while not self.task_queue.empty():
            current_work.append(self.task_queue.get())
            self.task_queue.task_done()
            # create a dictionary of the work indices for quick removal later
            if current_work[-1] is not None:
                work_order[i] = current_work[-1]
                dist_ids_work_order.add(current_work[-1].distributor_id)
            i += 1

        # no need to worry about race conditions because there are no state
        # changes in the FSM caused by this method

        # now update our internal state tracker
        self._update_internal_states()

        # now terminate any running jobs and remove from state tracker
        # for distributor_id in distributor_ids:
        for w in work_order.values():
            if w.distributor_id in distributor_ids:
                execution_pid = self._running_or_submitted.get(w.distributor_id)
                if execution_pid is not None:
                    # kill the process and remove it from the state tracker
                    parent = psutil.Process(execution_pid)
                    for child in parent.children(recursive=True):
                        child.kill()

        unexpected_distributor_ids = set(distributor_ids).difference(
            dist_ids_work_order
        )
        for distributor_id in unexpected_distributor_ids:
            logger.error(
                f"distributor_id {distributor_id} was requested to be terminated"
                " but is not submitted or running"
            )

        # if not running remove from queue and state tracker
        for index in sorted(work_order.keys(), reverse=True):
            w = work_order[index]
            if w.distributor_id in distributor_ids:
                del current_work[index]
                del self._running_or_submitted[w.distributor_id]

        # put remaining work back on queue
        for task in current_work:
            self.task_queue.put(task)

    def get_submitted_or_running(
        self, distributor_ids: Optional[List[str]] = None
    ) -> Set[str]:
        """Get tasks that are active."""
        self._update_internal_states()
        return set(self._running_or_submitted.keys())

    def submit_to_batch_distributor(
        self,
        command: str,
        name: str,
        requested_resources: Dict[str, Any],
    ) -> str:
        distributor_id = str(self._next_job_id)
        self._next_job_id += 1

        task = PickableTask(
            distributor_id, self.worker_node_entry_point + " " + command, "job"
        )
        self.task_queue.put(task)
        self._running_or_submitted.update({distributor_id: None})
        return distributor_id

    def submit_array_to_batch_distributor(
        self,
        command: str,
        name: str,
        requested_resources: Dict[str, Any],
        array_length: int,
    ) -> Dict[int, str]:
        """Submit an array task to the multiprocess cluster.

        Return: a mapping of array_step_id to distributor_id, output path, and error path
        """
        job_id = self._next_job_id
        self._next_job_id += 1

        mapping: Dict[int, str] = {}
        for array_step_id in range(0, array_length):
            distributor_id = self._get_subtask_id(job_id, array_step_id)
            mapping[array_step_id] = distributor_id

            task = PickableTask(
                distributor_id, self.worker_node_entry_point + " " + command, "array"
            )
            self.task_queue.put(task)
            self._running_or_submitted.update({distributor_id: None})

        return mapping

    def get_queueing_errors(self, distributor_ids: List[str]) -> Dict[str, str]:
        """Get the task instances that have errored out."""
        return {}

    def get_remote_exit_info(self, distributor_id: str) -> Tuple[str, str]:
        """Get the exit info about the task instance once it is done running."""
        raise RemoteExitInfoNotAvailable


class MultiprocessWorkerNode(ClusterWorkerNode):
    """Task instance info for an instance run with the Multiprocessing distributor."""

    def __init__(self) -> None:
        """Initialization of the multiprocess distributor worker node."""
        self._distributor_id: Optional[str] = None
        self._array_step_id: Optional[int] = None
        self._subtask_id: Optional[str] = None
        self._logfile_template = {
            "stdout": "{root}/{name}.o{job_id}",
            "stderr": "{root}/{name}.e{job_id}",
        }

    @property
    def distributor_id(self) -> Optional[str]:
        """The id from the distributor."""
        if self._distributor_id is None:
            jid = os.environ.get("JOB_ID")
            if jid:
                self._distributor_id = f"{jid}_{self.array_step_id}"
        return self._distributor_id

    def get_exit_info(self, exit_code: int, error_msg: str) -> Tuple[str, str]:
        """Exit code and message."""
        msg = f"Got exit_code: {exit_code}. Error message was: {error_msg}"
        return TaskInstanceStatus.ERROR, msg

    def get_usage_stats(self) -> Dict:
        """Usage information specific to the distributor."""
        return {}

    def initialize_logfile(self, log_type: str, log_dir: str, name: str) -> str:
        if log_dir:
            logpath = self._logfile_template[log_type].format(
                root=log_dir, name=name, job_id=self.distributor_id
            )
        else:
            logpath = "/dev/null"
        return logpath

    @property
    def array_step_id(self) -> Optional[int]:
        """Return array_step_id ."""
        if self._array_step_id is None:
            atid = os.environ.get("ARRAY_STEP_ID")
            if atid:
                self._array_step_id = int(atid)
        return self._array_step_id



"""The client for the Multiprocess executor."""

from typing import Dict, List, Tuple, Union

from jobmon.core.cluster_protocol import ClusterQueue


class MultiprocessQueue(ClusterQueue):
    """Implementation of the multiprocess executor queue, derived from ClusterQueue."""

    def __init__(self, queue_id: int, queue_name: str, parameters: Dict) -> None:
        """Intialization of the multiprocess queue.

        Get the limits from the database in the client.
        """
        self._queue_id = queue_id
        self._queue_name = queue_name
        self._parameters = parameters

    def validate_resources(
        self, strict: bool = False, **kwargs: Union[str, int, float]
    ) -> Tuple[bool, str]:
        """Ensure cores requested isn't more than available on that node."""
        is_valid = True
        msg = ""
        cores = kwargs.get("cores")
        core_parameters = self.parameters.get("cores")

        if core_parameters:
            min_cores, max_cores = core_parameters
        else:
            raise ValueError("min_cores and max_cores parameters not set on queue.")

        if cores:
            if cores > max_cores:
                msg += (
                    f"ResourceError: provided cores {cores} exceeds "
                    f"queue limit of {max_cores} "
                    f"for queue {self.queue_name}"
                )
                cores = max_cores
                if strict:
                    is_valid = False
            elif cores < min_cores:
                msg += (
                    f"ResourceError: provided cores {cores} is below "
                    f"queue minimum of {min_cores} "
                    f"for queue {self.queue_name}"
                )
                cores = min_cores
                if strict:
                    is_valid = False
        else:
            # Set cores to the queue minimum
            msg += f"Cores not provided, setting to {self.queue_name} minimum of {min_cores}"
            cores = min_cores
            if strict:
                is_valid = False
        return is_valid, msg

    def coerce_resources(self, **kwargs: Union[str, int, float]) -> Dict:
        return kwargs

    @property
    def queue_id(self) -> int:
        """Return the ID of the queue."""
        return self._queue_id

    @property
    def queue_name(self) -> str:
        """Return the name of the queue."""
        return self._queue_name

    @property
    def parameters(self) -> Dict:
        """Return the dictionary of parameters."""
        return self._parameters

    @property
    def required_resources(self) -> List:
        """No required resources specified for dummy executor, return empty list."""
        return []



"""Requester object to make HTTP requests to the Jobmon Flask services."""

from __future__ import annotations

import contextlib
import functools
import json
import logging
import logging.config
from typing import Any, Callable, Dict, Tuple, Type

import requests
import tenacity
import urllib3

from jobmon.core import __version__
from jobmon.core.configuration import JobmonConfig
from jobmon.core.exceptions import InvalidRequest, InvalidResponse

logger = logging.getLogger(__name__)

_OTEL_LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "otel": {
            "()": "jobmon.core.otlp.OpenTelemetryLogFormatter",
            "format": "%(asctime)s [%(levelname)s] [trace_id=%(trace_id)s,"
            " span_id=%(span_id)s, parent_span_id=%(parent_span_id)s]"
            " - %(message)s",
        },
    },
    "handlers": {
        "otel_text": {
            "level": "INFO",
            "class": "opentelemetry.sdk._logs.LoggingHandler",
            "formatter": "otel",
        },
    },
    "loggers": {
        "jobmon.core.requester": {
            "handlers": ["otel_text"],
            "level": "INFO",
            "propagate": False,
        },
    },
}


def http_request_ok(status_code: int) -> bool:
    """Return True if HTTP return codes that are deemed ok."""
    return status_code in (200, 302, 307)


class Requester(object):
    """Requester object to make HTTP requests to the Jobmon FastApi services."""

    # Class-level attribute to store the OtlpAPI instance
    _otlp_api = None

    def __init__(
        self,
        url: str,
        route_prefix: str = "",
        request_timeout: int = 20,
        retries_timeout: int = 300,
        retries_attempts: int = 10,
        use_otlp: bool = False,
    ) -> None:
        """Initialize the Requester object with the url to make requests to."""
        self.base_url = url
        self.route_prefix = route_prefix
        self.request_timeout = request_timeout
        self.retries_timeout = retries_timeout
        self.retries_attempts = retries_attempts
        if use_otlp and Requester._otlp_api is None:
            self._init_otlp()
        self.server_structlog_context: Dict[str, str] = {}

    @classmethod
    def _init_otlp(cls: Type[Requester]) -> None:
        from jobmon.core.otlp import OtlpAPI

        # setup connections to backend
        otlp_instance = OtlpAPI()
        otlp_instance.instrument_requests()

        logging.config.dictConfig(_OTEL_LOGGING_CONFIG)
        # setup tracer for Requester to use
        cls._otlp_api = otlp_instance

    @classmethod
    def from_defaults(cls: Type[Requester]) -> Requester:
        """Instantiate a requester from default config values."""
        config = JobmonConfig()
        service_url = config.get("http", "service_url")
        route_prefix = config.get("http", "route_prefix")
        request_timeout = config.get_int("http", "request_timeout")
        retries_timeout = config.get_int("http", "retries_timeout")
        retries_attempts = config.get_int("http", "retries_attempts")
        use_otlp = config.get_boolean("otlp", "http_enabled")
        return cls(
            service_url,
            route_prefix,
            request_timeout,
            retries_timeout,
            retries_attempts,
            use_otlp,
        )

    @property
    def url(self) -> str:
        """Return the base url for the requester."""
        return self.base_url + self.route_prefix

    def add_server_structlog_context(self, **kwargs: Any) -> None:
        """Add the structlogging context if it has been provided."""
        for key, value in kwargs.items():
            self.server_structlog_context[key] = value

    @contextlib.contextmanager
    def tracing_span(self, app_route: str, request_type: str) -> Any:
        if self._otlp_api:
            tracer = self._otlp_api.get_tracer("requester")
            with tracer.start_as_current_span("send_request") as span:
                span.set_attribute("http.method", request_type.upper())
                span.set_attribute("http.url", self.url + app_route)
                yield span
        else:
            yield None

    def _maybe_trace(self, func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Callable:
            with self.tracing_span(
                kwargs.get("app_route", "UNKNOWN"),
                kwargs.get("request_type", "UNKNOWN"),
            ):
                return func(*args, **kwargs)

        return wrapper

    def _maybe_retry(self, func: Callable, tenacious: bool) -> Any:
        if not tenacious:
            return func

        def should_retry_exception(exception: Any) -> Any:
            """Return True if we should retry on the given exception."""
            logger.warning(f"Exception occurred: {exception}")

            # Do not retry for certain client errors.
            if isinstance(exception, InvalidRequest):
                return False

            # Retry for specific exceptions.
            return isinstance(
                exception,
                (
                    InvalidResponse,
                    TimeoutError,
                    requests.ConnectionError,
                    requests.adapters.MaxRetryError,
                    requests.exceptions.ReadTimeout,
                    urllib3.exceptions.NewConnectionError,
                    urllib3.exceptions.MaxRetryError,
                ),
            )

        def raise_if_exceed_retry(retry_state: tenacity.RetryCallState) -> Any:
            """If we trigger retry error, raise informative RuntimeError."""
            # Check if the retry outcome is an exception
            outcome = retry_state.outcome
            if outcome and outcome.exception():
                exception = outcome.exception()
                raise RuntimeError(
                    f"Exceeded HTTP request retry budget due to: {exception}"
                ) from exception

        retrying = tenacity.retry(
            stop=(
                tenacity.stop_after_attempt(self.retries_attempts)
                | tenacity.stop_after_delay(self.retries_timeout)
            ),
            wait=tenacity.wait_exponential_jitter(initial=1, exp_base=2, jitter=1),
            retry=tenacity.retry_if_exception(should_retry_exception),
            retry_error_callback=raise_if_exceed_retry,
        )(func)

        return retrying

    def _send_request(
        self,
        app_route: str,
        message: dict,
        request_type: str,
    ) -> Tuple[int, Any]:
        # Construct URL
        route = self.url + app_route
        logger.info(f"Route: {route}, message: {message}")

        # Add version to query parameters
        params = {"client_jobmon_version": __version__}
        if request_type == "get":
            params.update(message)

        # Set headers, including the custom header for structlog context
        headers = {
            "Content-Type": "application/json",
            "X-Server-Structlog-Context": json.dumps(self.server_structlog_context),
        }

        # Send the appropriate request
        if request_type == "post":
            response = requests.post(
                route,
                params=params,
                json=message,
                headers=headers,
                timeout=self.request_timeout,
            )
        elif request_type == "get":
            response = requests.get(
                route,
                params=params,
                headers=headers,
                timeout=self.request_timeout,
            )
        elif request_type == "put":
            response = requests.put(
                route,
                params=params,
                json=message,
                headers=headers,
                timeout=self.request_timeout,
            )
        else:
            raise ValueError(
                f"request_type must be one of 'get', 'post', or 'put'. Got {request_type}"
            )

        # Extract status code and content
        status_code, content = get_content(response)

        # Raise the InvalidResponse exception based on the logic from should_retry_result
        if 499 < status_code < 600 or status_code == 423:
            raise InvalidResponse(
                f"Request failed due to status code {status_code} from {request_type.upper()} "
                f"request through route {app_route}. Response content: {content}"
            )

        # Keep the logic for other status codes that might be encountered but
        # aren't in the retry condition.
        if 400 <= status_code < 500:
            raise InvalidRequest(
                f"Client error with status code {status_code} from {request_type.upper()} "
                f"request through route {app_route}. Response content: {content}"
            )

        return status_code, content

    def send_request(
        self, app_route: str, message: dict, request_type: str, tenacious: bool = True
    ) -> Tuple[int, Any]:
        """Send a request to the Jobmon server."""

        def send_fn(
            app_route: str, message: dict, request_type: str
        ) -> Tuple[int, Any]:
            return self._send_request(app_route, message, request_type)

        send_method = self._maybe_retry(send_fn, tenacious)
        send_with_trace = self._maybe_trace(send_method)
        res = send_with_trace(
            app_route=app_route, message=message, request_type=request_type
        )

        return res


def get_content(response: Any) -> Tuple[int, Any]:
    """Parse the response."""
    if "application/json" in response.headers.get("Content-Type", ""):
        try:
            content = response.json()
        except TypeError:  # for test_client, response.json is a dict not fn
            content = response.json
    else:
        content = response.content
    return response.status_code, content



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Dict, List, Optional, Set, Tuple, Union

# the following try-except is to accommodate Python versions on both >=3.8 and 3.7.
# The Protocol was officially introduced in 3.8, with typing_extensions slapped on 3.7.
try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore

from jobmon.core import __version__
from jobmon.core.exceptions import RemoteExitInfoNotAvailable


class ClusterQueue(Protocol):
    """The protocol class for queues on a cluster."""

    @abstractmethod
    def __init__(self, queue_id: int, queue_name: str, parameters: Dict) -> None:
        """Initialization of ClusterQueue."""
        raise NotImplementedError

    @abstractmethod
    def validate_resources(
        self, strict: bool = False, **kwargs: Union[str, int, float]
    ) -> Tuple[bool, str]:
        """Ensures that requested resources aren't greater than what's available."""
        raise NotImplementedError

    @abstractmethod
    def coerce_resources(self, **kwargs: Union[str, int, float]) -> Dict:
        """Ensures that requested resources aren't greater than what's available."""
        raise NotImplementedError

    @property
    @abstractmethod
    def parameters(self) -> Dict:
        """Returns the dictionary of parameters."""
        raise NotImplementedError

    @property
    @abstractmethod
    def queue_name(self) -> str:
        """Returns the name of the queue."""
        raise NotImplementedError

    @property
    @abstractmethod
    def queue_id(self) -> int:
        """Returns the ID of the queue."""
        raise NotImplementedError

    @property
    @abstractmethod
    def required_resources(self) -> List:
        """Returns the list of resources that are required."""
        raise NotImplementedError


class ClusterDistributor(Protocol):
    """The protocol class for cluster distributors."""

    @abstractmethod
    def __init__(self, cluster_name: str, *args: str, **kwargs: str) -> None:
        """Initialization of ClusterQueue."""
        raise NotImplementedError

    @property
    @abstractmethod
    def worker_node_entry_point(self) -> str:
        """Path to jobmon worker_node_entry_point."""
        raise NotImplementedError

    @property
    @abstractmethod
    def cluster_name(self) -> str:
        """Return the name of the cluster type."""
        raise NotImplementedError

    @abstractmethod
    def start(self) -> None:
        """Start the distributor."""
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        """Stop the distributor."""
        raise NotImplementedError

    @abstractmethod
    def get_queueing_errors(self, distributor_ids: List[str]) -> Dict[str, str]:
        """Get the task instances that have errored out."""
        raise NotImplementedError

    @abstractmethod
    def get_submitted_or_running(
        self, distributor_ids: Optional[List[str]] = None
    ) -> Set[str]:
        """Check which task instances are active.

        Returns: a set strings
        """
        raise NotImplementedError

    @abstractmethod
    def terminate_task_instances(self, distributor_ids: List[str]) -> None:
        """Terminate task instances.

        If implemented, return a list of (task_instance_id, hostname) tuples for any
        task_instances that are terminated.
        """
        raise NotImplementedError

    @abstractmethod
    def get_remote_exit_info(self, distributor_id: str) -> Tuple[str, str]:
        """Get the exit info about the task instance once it is done running."""
        raise RemoteExitInfoNotAvailable

    @abstractmethod
    def submit_to_batch_distributor(
        self,
        command: str,
        name: str,
        requested_resources: Dict[str, Any],
    ) -> str:
        """Submit the command on the cluster technology and return a distributor_id.

        The distributor_id can be used to identify the associated TaskInstance, terminate
        it, monitor for missingness, or collect usage statistics. If an exception is raised by
        this method the task instance will move to "W" state and the exception will be logged
        in the database under the task_instance_error_log table.

        Args:
            command: command to be run
            name: name of task
            requested_resources: resource requests sent to distributor API

        Returns:
            A tuple indicating the distributor id, the full output file location,
            and full error location.
        """
        raise NotImplementedError

    def submit_array_to_batch_distributor(
        self,
        command: str,
        name: str,
        requested_resources: Dict[str, Any],
        array_length: int,
    ) -> Dict[int, str]:
        """Submit an array task to the underlying distributor and return a distributor_id.

        The distributor ID represents the ID of the overall array job, sub-tasks will have
        their own associated IDs.

        Args:
            command: the array worker node command to run
            name: name of the array
            requested_resources: resources with which to run the array
            array_length: how many tasks associated with the array
        Return:
            a mapping of array_step_id to distributor_id, output location, and error location.
        """
        raise NotImplementedError

    def build_worker_node_command(
        self,
        task_instance_id: Optional[int] = None,
        array_id: Optional[int] = None,
        batch_number: Optional[int] = None,
    ) -> str:
        """Build a command that can be executed by the worker_node.

        Args:
            task_instance_id: id for the given instance of this task
            array_id: id for the array if using an array strategy
            batch_number: if array strategy is used, the submission counter index to use

        Returns:
            (str) unwrappable command
        """
        wrapped_cmd = []
        if task_instance_id is not None:
            wrapped_cmd.extend(
                ["worker_node_job", "--task_instance_id", str(task_instance_id)]
            )
        elif array_id is not None and batch_number is not None:
            wrapped_cmd.extend(
                [
                    "worker_node_array",
                    "--array_id",
                    str(array_id),
                    "--batch_number",
                    str(batch_number),
                ]
            )
        else:
            raise ValueError(
                "Must specify either task_instance_id or array_id and batch_number. Got "
                f"task_instance_id={task_instance_id}, array_id={array_id}, "
                f"batch_number={batch_number}"
            )
        wrapped_cmd.extend(
            [
                "--expected_jobmon_version",
                __version__,
                "--cluster_name",
                self.cluster_name,
            ]
        )
        str_cmd = " ".join([str(i) for i in wrapped_cmd])
        return str_cmd


class ClusterWorkerNode(Protocol):
    """Base class defining interface for gathering executor info in the execution_wrapper.

    While not required, implementing get_usage_stats() will allow collection
    of CPU/memory utilization stats for each job.

    Get exit info is used to determine the error type if the task hits a
    system error of some variety.
    """

    @property
    @abstractmethod
    def distributor_id(self) -> Optional[str]:
        """Executor specific id assigned to a task instance."""
        raise NotImplementedError

    @abstractmethod
    def get_usage_stats(self) -> Dict:
        """Usage information specific to the exector."""
        raise NotImplementedError

    @abstractmethod
    def get_exit_info(self, exit_code: int, error_msg: str) -> Tuple[str, str]:
        """Error and exit code info from the executor."""
        raise NotImplementedError

    @abstractmethod
    def initialize_logfile(self, log_type: str, log_dir: str, name: str) -> str:
        """Error and exit code info from the executor."""
        raise NotImplementedError

    @property
    def array_step_id(self) -> Optional[int]:
        """The step id in each batch.

        For each array task instance, array_id, array_batch_num, and array_step_id
        should uniquely identify a subtask_id.

        It depends on the plug in whether you can generate the subtask_id using
        array_step_id.
        """
        raise NotImplementedError



from __future__ import annotations

import logging

from jobmon.core.constants import WorkflowRunStatus
from jobmon.core.requester import Requester

logger = logging.getLogger(__name__)


class DistributorWorkflowRun:
    """Implements workflow level bulk routes and tracks the in memory state on the distributor.

    when polling from the database we should work in task space and translate into array
    space in memory where appropriate.

    when pushing to the database we should work in CommandType (Workflow/Array/Task) space
    """

    def __init__(self, workflow_run_id: int, requester: Requester) -> None:
        """Initialization of DistributorWorkflowRun object."""
        self.workflow_run_id = workflow_run_id
        self.status = ""
        self.requester = requester

    def _update_status(self, status: str) -> None:
        """Update the status of the workflow_run with whatever status is passed."""
        app_route = f"/workflow_run/{self.workflow_run_id}/update_status"
        self.requester.send_request(
            app_route=app_route,
            message={"status": status},
            request_type="put",
        )
        self.status = status

    def transition_to_instantiated(self) -> None:
        self._update_status(WorkflowRunStatus.INSTANTIATED)

    def transition_to_launched(self) -> None:
        self._update_status(WorkflowRunStatus.LAUNCHED)

    def __hash__(self) -> int:
        """Returns the ID of the workflow run."""
        return self.workflow_run_id

    def __eq__(self, other: object) -> bool:
        """Check if the hashes of two WorkflowRuns are equivalent."""
        if not isinstance(other, DistributorWorkflowRun):
            return False
        else:
            return hash(self) == hash(other)

    def __lt__(self, other: DistributorWorkflowRun) -> bool:
        """Check if one hash is less than the hash of another."""
        return hash(self) < hash(other)



import importlib
from importlib.metadata import version
import pkgutil

__version__ = version("jobmon_core")


def _get_config_file_from_installer_plug_in() -> str:
    """Return the default configuration file path from the installer plug-in."""
    # if the installer plugin exists, use the config file form the plugin
    plugins = [
        plugin_name
        for finder, plugin_name, ispkg in pkgutil.iter_modules()
        if plugin_name.startswith("jobmon_installer")
    ]
    if len(plugins) == 1:
        plugin_name = plugins[0]
        module = importlib.import_module(plugin_name)
        get_config_file = getattr(module, "get_config_file")
        filepath = get_config_file()
        return filepath
    elif len(plugins) > 1:
        raise Exception(
            "Found multiple plugins while installing config from plugin, but only one "
            f'is allowed. Got "{plugins}"'
        )
    return ""


CONFIG_FILE_FROM_INSTALLER_PLUGIN = _get_config_file_from_installer_plug_in()



"""Custom Exceptions used throughout Jobmon."""


class ReturnCodes(object):
    """Bash return codes used in distributor wrapper."""

    OK = 0
    WORKER_NODE_ENV_FAILURE = 198
    WORKER_NODE_CLI_FAILURE = 199


class InvalidResponse(Exception):
    """Invalid Response type Exception."""

    pass


class InvalidRequest(Exception):
    """Invalid Request type Exception."""

    pass


class RemoteExitInfoNotAvailable(Exception):
    """Exception raised when Exit Info is not available for different executor types."""

    pass


class CallableReturnedInvalidObject(Exception):
    """Invalid Object got returned."""

    pass


class WorkflowAlreadyExists(Exception):
    """Workflow with the same workflow args already exists."""

    pass


class WorkflowAlreadyComplete(Exception):
    """This Workflow is already done."""

    pass


class WorkflowNotResumable(Exception):
    """This Workflow is not set to be resumed."""

    pass


class EmptyWorkflowError(Exception):
    """This Workflow is empty."""

    pass


class DistributorStartupTimeout(Exception):
    """Distributor was not able to start in time."""

    pass


class DistributorNotAlive(Exception):
    """The Distributor is not running."""

    pass


class DistributorUnexpected(Exception):
    """Unexpected situation in Distributor."""

    pass


class WorkflowRunStateError(Exception):
    """Error with the Workflow Run status."""

    pass


class ResumeSet(Exception):
    """Resume Exception."""

    pass


class NodeDependencyNotExistError(Exception):
    """Dependency does not exist."""

    pass


class DuplicateNodeArgsError(Exception):
    """Multiple nodes with the same args for the same TaskTemplate not allowed."""

    pass


class InvalidMemoryFormat(Exception):
    """Memory input invalid."""

    pass


class InvalidMemoryUnit(Exception):
    """Memory convert unit invalid."""

    pass


class ConfigError(Exception):
    """No configuration found for server."""

    pass


class InvalidStateTransition(Exception):
    """Invalid State Transition implementation."""

    def __init__(self, model: str, id: int, old_state: str, new_state: str) -> None:
        """Initialize Exception."""
        msg = f"Cannot transition {model} id: {id} from {old_state} to {new_state}"
        super(InvalidStateTransition, self).__init__(self, msg)


class TransitionError(Exception):
    """Transition failed."""

    pass


class WorkflowTestError(Exception):
    """Workflow Run encountered and error."""

    pass


class DistributorInterruptedError(Exception):
    """raised when signal is sent to distributor."""

    pass


class CyclicGraphError(Exception):
    """Cyclic graph detected."""

    pass



"""Start up distributing process."""

from typing import Optional

from jobmon.core.cluster import Cluster
from jobmon.core.requester import Requester
from jobmon.worker_node.worker_node_task_instance import WorkerNodeTaskInstance


class WorkerNodeFactory:
    def __init__(
        self, cluster_name: str, requester: Optional[Requester] = None
    ) -> None:
        """Initialization of the WorkerNode Factory."""
        self._cluster_name = cluster_name

        cluster = Cluster.get_cluster(cluster_name)
        self._worker_node_interface = cluster.get_worker_node()

    def get_job_task_instance(
        self,
        task_instance_id: int,
    ) -> WorkerNodeTaskInstance:
        """Set up and return WorkerNodeTaskInstance object."""
        worker_node_task_instance = WorkerNodeTaskInstance(
            cluster_interface=self._worker_node_interface,
            task_instance_id=task_instance_id,
        )
        return worker_node_task_instance

    def get_array_task_instance(
        self, array_id: int, batch_number: int
    ) -> WorkerNodeTaskInstance:
        """Set up and return WorkerNodeTaskInstance object."""
        requester = Requester.from_defaults()

        # Always assumed to be a value in the range [1, len(array)]
        array_step_id = self._worker_node_interface.array_step_id

        # Fetch from the database
        app_route = (
            f"/get_array_task_instance_id/{array_id}/{batch_number}/{array_step_id}"
        )
        _, resp = requester.send_request(
            app_route=app_route, message={}, request_type="get"
        )
        task_instance_id = resp["task_instance_id"]

        worker_node_task_instance = WorkerNodeTaskInstance(
            cluster_interface=self._worker_node_interface,
            task_instance_id=task_instance_id,
        )
        return worker_node_task_instance



"""Serializing data when going to and from the database."""

import ast
from datetime import datetime
import json
from typing import Any, Dict, List, Optional, Tuple, Union


class SerializeDistributorTask:
    """Serialize the data to and from the database for an DistributorTask object."""

    @staticmethod
    def to_wire(
        task_id: int,
        array_id: Optional[int],
        name: str,
        command: str,
        requested_resources: dict,
    ) -> tuple:
        """Submitting the above args to the database for an DistributorTask object."""
        return (task_id, array_id, name, command, requested_resources)

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Coerce types for all nullables that are cast using ast.literal_eval.

        It is a potential security issue but was the only solution I could find to turning
        the data into json twice.
        """
        return {
            "task_id": int(wire_tuple[0]),
            "array_id": int(wire_tuple[1]) if wire_tuple[1] is not None else None,
            "name": wire_tuple[2],
            "command": wire_tuple[3],
            "requested_resources": (
                {} if wire_tuple[4] is None else ast.literal_eval(wire_tuple[4])
            ),
        }


class SerializeSwarmTask:
    """Serialize the data to and from the db for a Swarm Task."""

    @staticmethod
    def to_wire(task_id: int, status: str) -> tuple:
        """Submit task id and status to the database from a SwarmTask object."""
        return task_id, status

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Get task id and status for a SwarmTask."""
        return {"task_id": int(wire_tuple[0]), "status": wire_tuple[1]}


class SerializeTaskInstance:
    """Serialize the data to and from the database for an DistributorTaskInstance."""

    @staticmethod
    def to_wire_distributor(
        task_instance_id: int,
        task_id: int,
        workflow_run_id: int,
        workflow_id: int,
        status: str,
        distributor_id: Union[str, None],
        task_resources_id: Optional[int] = None,
        array_id: Optional[int] = None,
        array_batch_num: Optional[int] = None,
        array_step_id: Optional[int] = None,
    ) -> tuple:
        """Submit the above args for an DistributorTaskInstance object to the database."""
        return (
            task_instance_id,
            task_id,
            workflow_run_id,
            workflow_id,
            status,
            distributor_id,
            task_resources_id,
            array_id,
            array_batch_num,
            array_step_id,
        )

    @staticmethod
    def kwargs_from_wire_distributor(wire_tuple: tuple) -> dict:
        """Retrieve the DistributorTaskInstance information from the database."""
        task_instance_id = int(wire_tuple[0])
        task_id = int(wire_tuple[1])
        workflow_run_id = int(wire_tuple[2])
        workflow_id = int(wire_tuple[3])
        status = str(wire_tuple[4])
        distributor_id = int(wire_tuple[5]) if wire_tuple[5] else None
        cluster_id = int(wire_tuple[6]) if wire_tuple[6] else None
        task_resources_id = int(wire_tuple[7]) if wire_tuple[7] else None
        array_id = int(wire_tuple[8]) if wire_tuple[8] else None
        array_batch_num = int(wire_tuple[9]) if wire_tuple[9] else None
        array_step_id = int(wire_tuple[10]) if wire_tuple[10] else None

        return {
            "task_instance_id": task_instance_id,
            "task_id": task_id,
            "workflow_run_id": workflow_run_id,
            "workflow_id": workflow_id,
            "status": status,
            "distributor_id": distributor_id,
            "cluster_id": cluster_id,
            "task_resources_id": task_resources_id,
            "array_id": array_id,
            "array_batch_num": array_batch_num,
            "array_step_id": array_step_id,
        }

    @staticmethod
    def to_wire_worker_node(
        task_instance_id: int,
        status: str,
        workflow_run_id: int,
        task_id: int,
        name: str,
        command: str,
        workflow_id: int,
        stdout_dir: str,
        stderr_dir: str,
    ) -> tuple:
        """Submit the above args for an DistributorTaskInstance object to the database."""
        return (
            task_instance_id,
            status,
            workflow_run_id,
            task_id,
            name,
            command,
            workflow_id,
            stdout_dir,
            stderr_dir,
        )

    @staticmethod
    def kwargs_from_wire_worker_node(wire_tuple: tuple) -> dict:
        """Retrieve the DistributorTaskInstance information from the database."""
        task_instance_id = int(wire_tuple[0])
        status = wire_tuple[1]
        workflow_run_id = int(wire_tuple[2])
        task_id = int(wire_tuple[3])
        name = wire_tuple[4]
        command = wire_tuple[5]
        workflow_id = int(wire_tuple[6])
        stdout_dir = wire_tuple[7]
        stderr_dir = wire_tuple[8]

        return {
            "task_instance_id": task_instance_id,
            "status": status,
            "workflow_run_id": workflow_run_id,
            "task_id": task_id,
            "name": name,
            "command": command,
            "workflow_id": workflow_id,
            "stdout_dir": stdout_dir,
            "stderr_dir": stderr_dir,
        }


class SerializeTaskInstanceErrorLog:
    """Serialize the data to and from the database for an TaskInstanceErrorLog."""

    @staticmethod
    def to_wire(
        task_instance_error_log_id: int, error_time: datetime, description: str
    ) -> tuple:
        """Submit the args for an SerializeTaskInstanceErrorLog object to the database."""
        return task_instance_error_log_id, str(error_time), description

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Retrieve the SerializeTaskInstanceErrorLog information from the database."""
        return {
            "task_instance_error_log_id": int(wire_tuple[0]),
            "error_time": str(wire_tuple[1]),
            "description": str(wire_tuple[2]),
        }


class SerializeExecutorTaskInstanceErrorLog:
    """Serialize the data to and from the database for an ExecutorTaskInstanceErrorLog."""

    @staticmethod
    def to_wire(
        task_instance_error_log_id: int, error_time: datetime, description: str
    ) -> tuple:
        """A to_wire method.

        Submit the above args for an SerializeExecutorTaskInstanceErrorLog
        object to the database.
        """
        return task_instance_error_log_id, error_time, description

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Retrieve the SerializeExecutorTaskInstanceErrorLog information from the database."""
        return {
            "task_instance_error_log_id": int(wire_tuple[0]),
            "error_time": str(wire_tuple[1]),
            "description": str(wire_tuple[2]),
        }


class SerializeClientTool:
    """Serialize the data to and from the database for a Tool object."""

    @staticmethod
    def to_wire(id: int, name: str) -> tuple:
        """Submit the id and name of a Tool to the database."""
        return (id, name)

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Get a Tool's information from the database."""
        return {"id": int(wire_tuple[0]), "name": wire_tuple[1]}


class SerializeClientToolVersion:
    """Serialize the data to and from the database for a ToolVersion."""

    @staticmethod
    def to_wire(id: int, tool_id: int) -> tuple:
        """Submit the id and tool_id for a Tool Version to the database."""
        return (id, tool_id)

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Get the Tool Version information from thee database."""
        return {"id": int(wire_tuple[0]), "tool_id": int(wire_tuple[1])}


class SerializeClientTaskTemplate:
    """Serialize the data to and from the database for a TaskTemplate."""

    @staticmethod
    def to_wire(id: int, tool_version_id: int, template_name: str) -> tuple:
        """Submit the TaskTemplate information for transfer over http."""
        return (id, tool_version_id, template_name)

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Convert packed wire format to kwargs for use in services."""
        return {
            "id": int(wire_tuple[0]),
            "tool_version_id": int(wire_tuple[1]),
            "template_name": wire_tuple[2],
        }


class SerializeClientTaskTemplateVersion:
    """Serialize the data to and from the database for a TaskTemplateVersion."""

    @staticmethod
    def to_wire(
        task_template_version_id: int,
        command_template: str,
        node_args: List[str],
        task_args: List[str],
        op_args: List[str],
        id_name_map: dict,
        task_template_id: int,
    ) -> Tuple[int, str, List[str], List[str], List[str], dict, int]:
        """Submit the TaskTemplateVersion information to the database."""
        return (
            task_template_version_id,
            command_template,
            node_args,
            task_args,
            op_args,
            id_name_map,
            task_template_id,
        )

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> Dict:
        """Get the TaskTemplateVersion info from the database."""
        return {
            "task_template_version_id": int(wire_tuple[0]),
            "command_template": wire_tuple[1],
            "node_args": wire_tuple[2],
            "task_args": wire_tuple[3],
            "op_args": wire_tuple[4],
            "id_name_map": wire_tuple[5],
            "task_template_id": wire_tuple[6],
        }


class SerializeWorkflowRun:
    """Serialize the data to and from the database for a WorkflowRun."""

    @staticmethod
    def to_wire(id: int, workflow_id: int) -> tuple:
        """Submit the WorkflowRun information to the database."""
        return (id, workflow_id)

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Get the WorkflowRun information from the database."""
        return {"id": int(wire_tuple[0]), "workflow_id": int(wire_tuple[1])}


class SerializeClusterType:
    """Serialize the data to and from the database for a ClusterType."""

    @staticmethod
    def to_wire(id: int, name: str, package_location: str) -> tuple:
        """Submit the ClusterType information to the database."""
        return (id, name, package_location)

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Get the Cluster information from the database."""
        return {
            "id": int(wire_tuple[0]),
            "name": str(wire_tuple[1]),
            "package_location": str(wire_tuple[2]),
        }


class SerializeCluster:
    """Serialize the data to and from the database for a Cluster."""

    @staticmethod
    def to_wire(
        id: int,
        name: str,
        cluster_type_name: str,
        connection_parameters: str,
    ) -> tuple:
        """Submit the Cluster information to the database."""
        return (id, name, cluster_type_name, connection_parameters)

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Get the Cluster information from the database."""
        connection_parameters = (
            json.loads(wire_tuple[3]) if wire_tuple[3] is not None else {}
        )
        return {
            "id": int(wire_tuple[0]),
            "name": str(wire_tuple[1]),
            "cluster_type_name": str(wire_tuple[2]),
            "connection_parameters": connection_parameters,
        }


class SerializeQueue:
    """Serialize the data to and from the database for a Queue."""

    @staticmethod
    def to_wire(id: int, name: str, parameters: str) -> tuple:
        """Submit the Queue information to the database."""
        return (id, name, parameters)

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Get the Queue information from the database."""
        return {
            "queue_id": int(wire_tuple[0]),
            "queue_name": str(wire_tuple[1]),
            "parameters": (
                {} if wire_tuple[2] is None else ast.literal_eval(wire_tuple[2])
            ),
        }


class SerializeTaskResourceUsage:
    """Serialize the data to and from the database for Task resource usage."""

    @staticmethod
    def to_wire(
        num_attempts: Optional[int] = None,
        nodename: Optional[str] = None,
        runtime: Optional[int] = None,
        memory: Optional[int] = None,
    ) -> tuple:
        """Submit the Task resource usage information to the database."""
        return num_attempts, nodename, runtime, memory

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Get the Task resource usage information from the database."""
        return {
            "num_attempts": wire_tuple[0],
            "nodename": wire_tuple[1],
            "runtime": wire_tuple[2],
            "memory": wire_tuple[3],
        }


class SerializeTaskTemplateResourceUsage:
    """Serialize the data to and from the database for TaskTemplate resource usage."""

    @staticmethod
    def to_wire(
        num_tasks: Optional[int] = None,
        min_mem: Optional[int] = None,
        max_mem: Optional[int] = None,
        mean_mem: Optional[float] = None,
        min_runtime: Optional[int] = None,
        max_runtime: Optional[int] = None,
        mean_runtime: Optional[float] = None,
        median_mem: Optional[float] = None,
        median_runtime: Optional[float] = None,
        ci_mem: Optional[List[Any]] = None,
        ci_runtime: Optional[List[Any]] = None,
    ) -> tuple:
        """Submit the TaskTemplate resource usage information to the database."""
        return (
            num_tasks,
            min_mem,
            max_mem,
            mean_mem,
            min_runtime,
            max_runtime,
            mean_runtime,
            median_mem,
            median_runtime,
            ci_mem,
            ci_runtime,
        )

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Get the TaskTemplate resource usage information from the database."""
        return {
            "num_tasks": wire_tuple[0],
            "min_mem": wire_tuple[1],
            "max_mem": wire_tuple[2],
            "mean_mem": wire_tuple[3],
            "min_runtime": wire_tuple[4],
            "max_runtime": wire_tuple[5],
            "mean_runtime": wire_tuple[6],
            "median_mem": wire_tuple[7],
            "median_runtime": wire_tuple[8],
            "ci_mem": wire_tuple[9],
            "ci_runtime": wire_tuple[10],
        }


class SerializeDistributorArray:
    """Serialize the data to and from the database for DistributorArray."""

    @staticmethod
    def to_wire(array_id: int, max_concurrently_running: int, name: str) -> tuple:
        """Submit the TaskTemplate resource usage information to the database."""
        return array_id, max_concurrently_running, name

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Get the TaskTemplate resource usage information from the database."""
        return {
            "array_id": wire_tuple[0],
            "max_concurrently_running": wire_tuple[1],
            "name": wire_tuple[2],
        }


class SerializeDistributorWorkflow:
    """Serialize the data to and from the database for DistributorWorkflow."""

    @staticmethod
    def to_wire(workflow_id: int, dag_id: int, max_concurrently_running: int) -> tuple:
        """Serialize the workflow metadata used in the distributor."""
        return (
            workflow_id,
            dag_id,
            max_concurrently_running,
        )

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Deserialize the workflow metadata used in the distributor."""
        return {"workflow_id": wire_tuple[0], "max_concurrently_running": wire_tuple[1]}


class SerializeTaskResources:
    """Serialize the data to and from the database for Task resources."""

    @staticmethod
    def to_wire(
        task_resources_id: int,
        queue_name: str,
        task_resources_type_id: str,
        requested_resources: str,
    ) -> tuple:
        """Serialize the TaskResources metadata."""
        return (
            task_resources_id,
            queue_name,
            task_resources_type_id,
            requested_resources,
        )

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Deserialize the TaskResources metadata."""
        return {
            "task_resources_id": wire_tuple[0],
            "queue_name": wire_tuple[1],
            "task_resources_type_id": wire_tuple[2],
            "requested_resources": wire_tuple[3],
        }


class SerializeTaskInstanceBatch:
    """Serialize the data to and from the database for TaskInstance batch."""

    @staticmethod
    def to_wire(
        array_id: int,
        array_name: str,
        array_batch_num: int,
        task_resources_id: int,
        task_instance_ids: List[int],
    ) -> tuple:
        """Serialize the TaskInstanceBatch metadata."""
        return (
            array_id,
            array_name,
            array_batch_num,
            task_resources_id,
            task_instance_ids,
        )

    @staticmethod
    def kwargs_from_wire(wire_tuple: tuple) -> dict:
        """Deserialize the TaskInstanceBatch metadata."""
        return {
            "array_id": wire_tuple[0],
            "array_name": wire_tuple[1],
            "array_batch_num": wire_tuple[2],
            "task_resources_id": wire_tuple[3],
            "task_instance_ids": wire_tuple[4],
        }



import os
import subprocess
from typing import Optional, Union


def true_path(
    file_or_dir: Optional[Union[bytes, str]] = None, executable: Optional[str] = None
) -> str:
    """Get true path to file or executable.

    Args:
        file_or_dir (str): partial file path, to be expanded as per the current
            user
        executable (str): the name of an executable, which will be resolved
            using "which"

    Specify one of the two arguments, not both.
    """
    if file_or_dir is not None:
        f = file_or_dir
    elif executable is not None:
        f = subprocess.check_output(["which", str(executable)])
    else:
        raise ValueError("true_path: file_or_dir and executable " "cannot both be null")

    # Be careful, in python 3 check_output returns bytes
    if not isinstance(f, str):
        f = f.decode("utf-8")
    f = os.path.abspath(os.path.expanduser(f))
    return f.strip(" \t\r\n")



from __future__ import annotations

import logging
from typing import Any, Callable

logger = logging.getLogger(__name__)


class DistributorCommand:
    def __init__(self, func: Callable[..., None], *args: Any, **kwargs: Any) -> None:
        """A command to be run by the distributor service.

        Args:
            func: a callable which does work and optionally modifies task instance state.
            *args: positional args to be passed into func.
            **kwargs: kwargs to be passed into func.
        """
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self.error_raised = False

    def __call__(self, raise_on_error: bool = False) -> None:
        try:
            self._func(*self._args, **self._kwargs)
        except Exception as e:
            if raise_on_error:
                raise
            else:
                logger.warning(e)
